"""
This module is largely inspired by django-rest-framework settings.

Settings for the SCIM Service Provider are all namespaced in the
SCIM_SERVICE_PROVIDER setting. For example your project's `settings.py`
file might look like this:

SCIM_SERVICE_PROVIDER = {
    'USER_ADAPTER': 'django_scim.adapters.SCIMUser',
    'GROUP_ADAPTER': 'django_scim.adapters.SCIMGroup',
}

This module provides the `scim_settings` object, that is used to access
SCIM Service Provider settings, checking for user settings first, then falling
back to the defaults.
"""
from __future__ import unicode_literals
from six import string_types

import importlib

from django.conf import settings


# Settings defined by user in root settings file for their project.
USER_SETTINGS = getattr(settings, 'SCIM_SERVICE_PROVIDER', None)

DEFAULTS = {
    'USER_ADAPTER': 'django_scim.adapters.SCIMUser',
    'GROUP_MODEL': 'django.contrib.auth.models.Group',
    'GROUP_ADAPTER': 'django_scim.adapters.SCIMGroup',
    'USER_FILTER_PARSER': 'django_scim.filters.UserFilterQuery',
    'GROUP_FILTER_PARSER': 'django_scim.filters.GroupFilterQuery',
    'SERVICE_PROVIDER_CONFIG_MODEL': 'django_scim.models.SCIMServiceProviderConfig',
    'BASE_LOCATION_GETTER': 'django_scim.utils.default_base_scim_location_getter',
    'GET_EXTRA_MODEL_FILTER_KWARGS_GETTER': 'django_scim.utils.default_get_extra_model_filter_kwargs_getter',
    'GET_EXTRA_MODEL_EXCLUDE_KWARGS_GETTER': 'django_scim.utils.default_get_extra_model_exclude_kwargs_getter',
    'SCHEMAS_GETTER': 'django_scim.schemas.default_schemas_getter',
    'DOCUMENTATION_URI': None,
    'SCHEME': 'https',
    'NETLOC': None,
    'AUTHENTICATION_SCHEMES': [],
    'WWW_AUTHENTICATE_HEADER': 'Basic realm="django-scim2"',
}

# List of settings that cannot be empty
MANDATORY = (
    'NETLOC',
    'AUTHENTICATION_SCHEMES',
)

# List of settings that may be in string import notation.
IMPORT_STRINGS = (
    'USER_ADAPTER',
    'GROUP_MODEL',
    'GROUP_ADAPTER',
    'USER_FILTER_PARSER',
    'GROUP_FILTER_PARSER',
    'SERVICE_PROVIDER_CONFIG_MODEL',
    'BASE_LOCATION_GETTER',
    'GET_EXTRA_MODEL_FILTER_KWARGS_GETTER',
    'GET_EXTRA_MODEL_EXCLUDE_KWARGS_GETTER',
    'SCHEMAS_GETTER',
)


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if isinstance(val, string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        parts = val.split('.')
        module_path, class_name = '.'.join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError as e:
        msg = "Could not import '%s' for setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class SCIMServiceProviderSettings(object):
    """
    A settings object, that allows SCIM Service Provider settings to be accessed as properties.

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, user_settings=None, defaults=None, import_strings=None, mandatory=None):
        self.user_settings = user_settings or {}
        self.defaults = defaults or {}
        self.import_strings = import_strings or ()
        self.mandatory = mandatory or ()

    def __getattr__(self, attr):
        if attr not in self.defaults.keys():
            raise AttributeError("Invalid SCIMServiceProvider setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if val and attr in self.import_strings:
            val = perform_import(val, attr)

        self.validate_setting(attr, val)

        # Cache the result
        setattr(self, attr, val)
        return val

    def validate_setting(self, attr, val):
        if not val and attr in self.mandatory:
            raise AttributeError("SCIMServiceProvider setting: '%s' is mandatory" % attr)


scim_settings = SCIMServiceProviderSettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS, MANDATORY)




import os
import json


SCHEMAS_DIR = os.path.dirname(os.path.abspath(__file__))
SCHEMA_SUB_DIRS = ('core', 'extension')


def load_schemas():
    schemas = []
    for dir_ in SCHEMA_SUB_DIRS:
        sub_dir = os.path.join(SCHEMAS_DIR, dir_)
        files = os.listdir(sub_dir)
        files = [os.path.join(sub_dir, f) for f in files if f.lower().endswith('.json')]
        for file_ in files:
            with open(file_) as fp:
                schemas.append(json.load(fp))

    return schemas


ALL = load_schemas()


def default_schemas_getter():
    return ALL



"""
Adapters are used to convert the data model described by the SCIM 2.0
specification to a data model that fits the data provided by the application
implementing a SCIM api.

For example, in a Django app, there are User and Group models that do
not have the same attributes/fields that are defined by the SCIM 2.0
specification. The Django User model has both ``first_name`` and ``last_name``
attributes but the SCIM speicifcation requires this same data be sent under
the names ``givenName`` and ``familyName`` respectively.

An adapter is instantiated with a model instance. Eg::

    user = get_user_model().objects.get(id=1)
    scim_user = SCIMUser(user)
    ...

"""
from six.moves.urllib.parse import urljoin

from django.contrib.auth import get_user_model
from django.urls import reverse
from django import core

from . import constants
from . import exceptions
from .utils import get_base_scim_location_getter
from .utils import get_group_adapter
from .utils import get_user_adapter


class SCIMMixin(object):
    def __init__(self, obj, request=None):
        self.obj = obj
        self._request = request

    @property
    def request(self):
        if self._request:
            return self._request

        raise RuntimeError('Adapter is not associated with a request object. '
                           'Set object.request to avoid this error.')

    @request.setter
    def request(self, value):
        self._request = value

    @property
    def id(self):
        return str(self.obj.scim_id)

    @property
    def path(self):
        return reverse(self.url_name, kwargs={'uuid': self.obj.scim_id})

    @property
    def location(self):
        return urljoin(get_base_scim_location_getter()(self.request), self.path)

    def save(self):
        self.obj.save()

    def delete(self):
        self.obj.__class__.objects.filter(id=self.obj.scim_id).delete()

    def handle_operations(self, operations):
        """
        The SCIM specification allows for making changes to specific attributes
        of a model. These changes are sent in PATCH requests and are batched into
        operations to be performed on a object. Operations can have an op code
        of 'add', 'remove', or 'replace'. This method iterates through all of the
        operations in ``operations`` and calls the appropriate handler (defined
        on the appropriate adapter) for each.
        """
        for operation in operations:
            path = operation.get('path')
            value = operation.get('value')

            path, value = self.parse_path_and_value(path, value)

            op_code = operation.get('op').lower()
            op_code = 'handle_' + op_code
            handler = getattr(self, op_code)

            handler(path, value, operation)

    def parse_path_and_value(self, path, value):
        """
        Return new path and value given an original path and value.

        This method can be overridden to provide a more usable path and value within the
        associated handle methods.
        """
        return path, value

    def handle_add(self, path, value, operation):
        raise exceptions.NotImplementedError

    def handle_remove(self, path, value, operation):
        raise exceptions.NotImplementedError

    def handle_replace(self, path, value, operation):
        raise exceptions.NotImplementedError


class SCIMUser(SCIMMixin):
    """
    Adapter for adding SCIM functionality to a Django User object.

    This adapter can be overriden; see the ``USER_ADAPTER`` setting
    for details.
    """
    # not great, could be more decoupled. But \__( )__/ whatevs.
    url_name = 'scim:users'
    resource_type = 'User'

    @property
    def display_name(self):
        """
        Return the displayName of the user per the SCIM spec.
        """
        if self.obj.first_name and self.obj.last_name:
            return u'{0.first_name} {0.last_name}'.format(self.obj)
        return self.obj.username

    @property
    def name_formatted(self):
        return self.display_name

    @property
    def emails(self):
        """
        Return the email of the user per the SCIM spec.
        """
        return [{'value': self.obj.email, 'primary': True}]

    @property
    def groups(self):
        """
        Return the groups of the user per the SCIM spec.
        """
        group_qs = self.obj.scim_groups.all()
        scim_groups = [get_group_adapter()(g, self.request) for g in group_qs]

        dicts = []
        for group in scim_groups:
            d = {
                'value': group.id,
                '$ref': group.location,
                'display': group.display_name,
            }
            dicts.append(d)

        return dicts

    @property
    def meta(self):
        """
        Return the meta object of the user per the SCIM spec.
        """
        d = {
            'resourceType': self.resource_type,
            'created': self.obj.date_joined.isoformat(),
            'lastModified': self.obj.date_joined.isoformat(),
            'location': self.location,
        }

        return d

    def to_dict(self):
        """
        Return a ``dict`` conforming to the SCIM User Schema,
        ready for conversion to a JSON object.
        """
        d = {
            'schemas': [constants.SchemaURI.USER],
            'id': self.id,
            'userName': self.obj.username,
            'name': {
                'givenName': self.obj.first_name,
                'familyName': self.obj.last_name,
                'formatted': self.name_formatted,
            },
            'displayName': self.display_name,
            'emails': self.emails,
            'active': self.obj.is_active,
            'groups': self.groups,
            'meta': self.meta,
        }

        return d

    def from_dict(self, d):
        """
        Consume a ``dict`` conforming to the SCIM User Schema, updating the
        internal user object with data from the ``dict``.

        Please note, the user object is not saved within this method. To
        persist the changes made by this method, please call ``.save()`` on the
        adapter. Eg::

            scim_user.from_dict(d)
            scim_user.save()
        """
        username = d.get('userName')
        self.obj.username = username or ''

        first_name = d.get('name', {}).get('givenName')
        self.obj.first_name = first_name or ''

        last_name = d.get('name', {}).get('familyName')
        self.obj.last_name = last_name or ''

        emails = d.get('emails', [])
        primary_emails = [e['value'] for e in emails if e.get('primary')]
        emails = primary_emails + emails
        email = emails[0] if emails else None
        self.obj.email = email

        cleartext_password = d.get('password')
        if cleartext_password:
            self.obj.set_password(cleartext_password)

        active = d.get('active')
        if active is not None:
            self.obj.is_active = active

    @classmethod
    def resource_type_dict(cls, request=None):
        """
        Return a ``dict`` containing ResourceType metadata for the user object.
        """
        id_ = cls.resource_type
        path = reverse('scim:resource-types', kwargs={'uuid': id_})
        location = urljoin(get_base_scim_location_getter()(request), path)
        return {
            'schemas': [constants.SchemaURI.RESOURCE_TYPE],
            'id': id_,
            'name': 'User',
            'endpoint': reverse('scim:users'),
            'description': 'User Account',
            'schema': constants.SchemaURI.USER,
            'meta': {
                'location': location,
                'resourceType': 'ResourceType'
            }
        }

    def handle_replace(self, path, value, operation):
        """
        Handle the replace operations.
        """
        attr_map = {
            'familyName': 'last_name',
            'givenName': 'first_name',
            'userName': 'username',
            'active': 'is_active',
        }

        attrs = value or {}

        for attr, attr_value in attrs.items():
            if attr in attr_map:
                setattr(self.obj, attr_map.get(attr), attr_value)
            elif attr == 'emails':
                primary_emails = [e for e in attr_value if e.get('primary')]
                if primary_emails:
                    email = primary_emails[0].get('value')
                elif attr_value:
                    email = attr_value[0].get('value')
                else:
                    raise exceptions.BadRequestError('Invalid email value')

                try:
                    validator = core.validators.EmailValidator()
                    validator(email)
                except core.exceptions.ValidationError:
                    raise exceptions.BadRequestError('Invalid email value')

                self.obj.email = email

            else:
                raise exceptions.NotImplementedError('Not Implemented')

        self.obj.save()


class SCIMGroup(SCIMMixin):
    """
    Adapter for adding SCIM functionality to a Django Group object.

    This adapter can be overriden; see the ``GROUP_ADAPTER``
    setting for details.
    """
    # not great, could be more decoupled. But \__( )__/ whatevs.
    url_name = 'scim:groups'
    resource_type = 'Group'

    @property
    def display_name(self):
        """
        Return the displayName of the group per the SCIM spec.
        """
        return self.obj.name

    @property
    def members(self):
        """
        Return a list of user dicts (ready for serialization) for the members
        of the group.

        :rtype: list
        """
        users = self.obj.user_set.all()
        scim_users = [get_user_adapter()(user, self.request) for user in users]

        dicts = []
        for user in scim_users:
            d = {
                'value': user.id,
                '$ref': user.location,
                'display': user.display_name,
            }
            dicts.append(d)

        return dicts

    @property
    def meta(self):
        """
        Return the meta object of the group per the SCIM spec.
        """
        d = {
            'resourceType': self.resource_type,
            'location': self.location,
        }

        return d

    def to_dict(self):
        """
        Return a ``dict`` conforming to the SCIM Group Schema,
        ready for conversion to a JSON object.
        """
        return {
            'schemas': [constants.SchemaURI.GROUP],
            'id': self.id,
            'displayName': self.display_name,
            'members': self.members,
            'meta': self.meta,
        }

    def from_dict(self, d):
        """
        Consume a ``dict`` conforming to the SCIM Group Schema, updating the
        internal group object with data from the ``dict``.

        Please note, the group object is not saved within this method. To
        persist the changes made by this method, please call ``.save()`` on the
        adapter. Eg::

            scim_group.from_dict(d)
            scim_group.save()
        """
        name = d.get('displayName')
        self.obj.name = name or ''

    @classmethod
    def resource_type_dict(cls, request=None):
        """
        Return a ``dict`` containing ResourceType metadata for the group object.
        """
        id_ = cls.resource_type
        path = reverse('scim:resource-types', kwargs={'uuid': id_})
        location = urljoin(get_base_scim_location_getter()(request), path)
        return {
            'schemas': [constants.SchemaURI.RESOURCE_TYPE],
            'id': id_,
            'name': 'Group',
            'endpoint': reverse('scim:groups'),
            'description': 'Group',
            'schema': constants.SchemaURI.GROUP,
            'meta': {
                'location': location,
                'resourceType': 'ResourceType'
            }
        }

    def handle_add(self, path, value, operation):
        """
        Handle add operations.
        """
        if path == 'members':
            members = value or []
            ids = [int(member.get('value')) for member in members]
            users = get_user_model().objects.filter(id__in=ids)

            if len(ids) != users.count():
                raise exceptions.BadRequestError('Can not add a non-existent user to group')

            for user in users:
                self.obj.user_set.add(user)

        else:
            raise exceptions.NotImplementedError

    def handle_remove(self, path, value, operation):
        """
        Handle remove operations.
        """
        if path == 'members':
            members = value or []
            ids = [int(member.get('value')) for member in members]
            users = get_user_model().objects.filter(id__in=ids)

            if len(ids) != users.count():
                raise exceptions.BadRequestError('Can not remove a non-existent user from group')

            for user in users:
                self.obj.user_set.remove(user)

        else:
            raise exceptions.NotImplementedError

    def handle_replace(self, path, value, operation):
        """
        Handle the replace operations.
        """
        if path == 'name':
            name = value[0].get('value')
            self.obj.name = name
            self.obj.save()

        else:
            raise exceptions.NotImplementedError




try:
    from django.urls import re_path
except ImportError:
    from django.conf.urls import url as re_path

from . import views


app_name = 'scim'

urlpatterns = [
    # This endpoint is used soley for middleware url purposes.
    re_path(r'^$',
            views.SCIMView.as_view(implemented=False),
            name='root'),

    re_path(r'^.search$',
            views.SearchView.as_view(implemented=False),
            name='search'),

    re_path(r'^Users/.search$',
            views.UserSearchView.as_view(),
            name='users-search'),

    re_path(r'^Users(?:/(?P<uuid>[^/]+))?$',
            views.UsersView.as_view(),
            name='users'),

    re_path(r'^Groups/.search$',
            views.GroupSearchView.as_view(),
            name='groups-search'),

    re_path(r'^Groups(?:/(?P<uuid>[^/]+))?$',
            views.GroupsView.as_view(),
            name='groups'),

    re_path(r'^Me$',
            views.SCIMView.as_view(implemented=False),
            name='me'),

    re_path(r'^ServiceProviderConfig$',
            views.ServiceProviderConfigView.as_view(),
            name='service-provider-config'),

    re_path(r'^ResourceTypes(?:/(?P<uuid>[^/]+))?$',
            views.ResourceTypesView.as_view(),
            name='resource-types'),

    re_path(r'^Schemas(?:/(?P<uuid>[^/]+))?$',
            views.SchemasView.as_view(),
            name='schemas'),

    re_path(r'^Bulk$',
            views.SCIMView.as_view(implemented=False),
            name='bulk'),
]



"""
Transform filter query into QuerySet
"""
from django.contrib.auth import get_user_model

from scim2_filter_parser.query import Query

from .utils import get_group_model


class FilterQuery:
    model_getter = None
    joins = ()
    attr_map = None
    query_class = Query

    @classmethod
    def table_name(cls):
        return cls.model_getter()._meta.db_table

    @classmethod
    def search(cls, filter_query, request=None):
        q = cls.query_class(filter_query, cls.table_name(), cls.attr_map, cls.joins)
        if q.where_sql is None:
            return cls.model_getter().objects.none()

        return cls.model_getter().objects.raw(q.sql, q.params)


class UserFilterQuery(FilterQuery):
    model_getter = get_user_model
    attr_map = {}


class GroupFilterQuery(FilterQuery):
    model_getter = get_group_model
    attr_map = {}




#!/usr/bin/env python
# Installs django-scim2

from setuptools import setup
import os
import sys
import unittest
import sys


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def get_version():
    path = os.path.join(BASE_DIR, 'django_scim', '__init__.py')
    version_line = None
    with open(path) as fp:
        for line in fp:
            if line.startswith('__version__'):
                version_line = line
                break
    if not version_line:
        raise ValueError('Unable to find version line in __init__.py')

    version = version_line.split('=')[1].strip().strip("'")
    return version


def long_description():
    """Get the long description from the README"""
    return open(os.path.join(BASE_DIR, 'README.rst')).read()


def run_tests():
    os.environ['DJANGO_SETTINGS_MODULE'] = os.environ.get('DJANGO_SETTINGS_MODULE', 'test_settings')

    import django
    django.setup()
    from django.test.utils import get_runner
    from django.conf import settings

    # eg. TEST_FILTER=django_scim.tests.test_views.UserTestCase.test_get_user_by_id
    test_filter = os.environ.get('TEST_FILTER')
    test_labels = [test_filter] if test_filter else []

    test_runner = get_runner(settings)
    failures = test_runner(
        pattern='test_*.py',
        verbosity=1,
        interactive=True
    ).run_tests(test_labels)
    sys.exit(failures)


setup(
    name='django-scim2',
    version=get_version(),
    description='A partial implementation of the SCIM 2.0 provider specification for use with Django.',
    url='https://github.com/15five/django-scim2',
    download_url='https://github.com/15five/django-scim2/archive/master.zip',
    maintainer='Paul Logston',
    maintainer_email='paul@15five.com',
    author='Paul Logston',
    author_email='paul@15five.com',
    keywords='django scim 2.0',
    license='MIT',
    long_description=long_description(),
    packages=['django_scim'],
    install_requires=[
        'python-dateutil>=2.6.0',
        'six',
        'scim2-filter-parser==0.1.0',
    ],
    tests_require=[
        'mock',
    ],
    test_suite='setup.run_tests',
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)

